package com.appiancorp.ps.sail.expr;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Logger;

import com.appiancorp.ps.sail.type.PortalReport;
import com.appiancorp.ps.sail.type.PortalReportCell;
import com.appiancorp.ps.sail.type.PortalReportColumn;
import com.appiancorp.ps.sail.type.PortalReportColumnData;
import com.appiancorp.ps.sail.type.PortalReportDataSubset;
import com.appiancorp.ps.sail.type.PortalReportFilter;
import com.appiancorp.services.ServiceContext;
import com.appiancorp.suiteapi.common.Constants;
import com.appiancorp.suiteapi.common.LocalObject;
import com.appiancorp.suiteapi.common.ObjectTypeMapping;
import com.appiancorp.suiteapi.common.exceptions.ExpressionException;
import com.appiancorp.suiteapi.common.exceptions.PrivilegeException;
import com.appiancorp.suiteapi.common.paging.PagingInfo;
import com.appiancorp.suiteapi.common.paging.SortInfo;
import com.appiancorp.suiteapi.expression.annotations.AppianScriptingFunctionsCategory;
import com.appiancorp.suiteapi.expression.annotations.Function;
import com.appiancorp.suiteapi.expression.annotations.Parameter;
import com.appiancorp.suiteapi.knowledge.DocumentDataType;
import com.appiancorp.suiteapi.process.analytics2.Column;
import com.appiancorp.suiteapi.process.analytics2.Filter;
import com.appiancorp.suiteapi.process.analytics2.ProcessAnalyticsService;
import com.appiancorp.suiteapi.process.analytics2.ProcessReport;
import com.appiancorp.suiteapi.process.analytics2.ReportData;
import com.appiancorp.suiteapi.process.analytics2.ReportDisplay;
import com.appiancorp.suiteapi.process.analytics2.ReportResultPage;
import com.appiancorp.suiteapi.process.analytics2.SimpleColumnFilter;
import com.appiancorp.suiteapi.process.exceptions.ReportComplexityException;
import com.appiancorp.suiteapi.process.exceptions.ReportSizeException;
import com.appiancorp.suiteapi.process.exceptions.UnsupportedReportSpecificationException;
import com.appiancorp.suiteapi.type.NamedTypedValue;
import com.appiancorp.suiteapi.type.TypeService;
import com.appiancorp.suiteapi.type.TypedValue;
import com.appiancorp.tools.cdth.CDTHelper;
import com.appiancorp.tools.cdth.CDTHelperUtils;

@AppianScriptingFunctionsCategory
public class PortalReport2SAIL {
	private static final Logger LOG = Logger.getLogger(PortalReport2SAIL.class);

	@Function
	public TypedValue executePortalReportForSail(
		ServiceContext sc,
		ProcessAnalyticsService pas,
		TypeService ts,
		@Parameter @DocumentDataType Long reportId,
		@Parameter(required = false) Long[] contextIds,
		@Parameter(required = false) PortalReportFilter[] filters
	) throws PrivilegeException, UnsupportedReportSpecificationException, ReportComplexityException, ReportSizeException, ExpressionException {
		ProcessReport processReport = pas.getProcessReport(reportId);
		ReportData reportData = getReportDataWithContext(sc, processReport, contextIds, filters);
		ReportResultPage rp = pas.getReportPageWithTypedValues(reportData);

		switch(processReport.getDisplay().getDisplayType()) {
			case ReportDisplay.DISPLAY_PIE_CHART:
				return getChartReport(ts, processReport, reportData, rp, "PieChart");
			default:
				return getChartReport(ts, processReport, reportData, rp, "StackedBarChart");
		}
	}

	private TypedValue getChartReport(TypeService ts, ProcessReport processReport, ReportData reportData, ReportResultPage rp, String typeName) {
		CDTHelper chart = CDTHelperUtils.fromQualifiedName(ts, "http://www.appian.com/ae/types/2009", typeName);

		chart.setValue("label", processReport.getDisplay().getName());
		chart.setValue("instructions", processReport.getDisplay().getDescription());
		chart.setValue("showDataLabels", processReport.getDisplay().getShowLegend());

		for (HashMap<String, Object> map : (HashMap[]) rp.getResults()) {
			// Always use StackedBarChartSeries, Appian will cast it to the correct type.
			CDTHelper series = CDTHelperUtils.fromQualifiedName(ts, "http://www.appian.com/ae/types/2009", "StackedBarChartSeries");

			for (Column column : reportData.getColumns()) {
				if (column.getShow()) {
					if (column.getFormatToken().equals("string")) {
						series.setValue(
							"label",
							((TypedValue) map.get("c" + column.getLocalId())).getValue()
						);
					} else if (column.getFormatToken().equals("number")) {
						Object valObj = ((TypedValue) map.get("c" + column.getLocalId())).getValue();
						Double value = null;

						if (valObj == null) {
							value = 0d;
						} else if (valObj instanceof Long) {
							value = ((Long) valObj).doubleValue();
						} else if (valObj instanceof Double) {
							value = (Double) valObj;
						}

						series.addValue(
							"data",
							value
						);
					}
				}
			}

			chart.addValue("series", series);
		}

		return chart.toTypedValue();
	}

	@Function
	public PortalReportDataSubset getPortalReportDataSubset(
		ServiceContext sc,
		ProcessAnalyticsService pas,
		TypeService ts,
		@Parameter @DocumentDataType Long reportId,
		@Parameter(required = false) PagingInfo pagingInfo,
		@Parameter(required = false) Long[] contextIds,
		@Parameter(required = false) PortalReportFilter[] filters
	) throws PrivilegeException, UnsupportedReportSpecificationException, ReportComplexityException, ReportSizeException, ExpressionException {
		ProcessReport processReport = null;

		try {			
			processReport = pas.getProcessReport(reportId);			
		} catch (Exception e) {
			LOG.warn("Unable to find report with report id: " + reportId);
			return null;
		}

		ReportData reportData = getReportDataWithContext(sc, processReport, contextIds, filters);

		if (pagingInfo == null) {
			// use report settings
			pagingInfo = new PagingInfo(
				reportData.getStartIndex() + 1,  // paging infos use 1 based indexes
				reportData.getBatchSize()
			);
		} else {
			reportData.setStartIndex(pagingInfo.getStartIndex() - 1); // reports are zero based indexes
			reportData.setBatchSize(pagingInfo.getBatchSize());

			for (SortInfo sortInfo : pagingInfo.getSort()) {
				reportData.setSortColumnLocalId(Integer.parseInt(sortInfo.getField()));
				reportData.setSortOrder(sortInfo.isAscending() ? Constants.SORT_ORDER_ASCENDING : Constants.SORT_ORDER_DESCENDING);
			}
		}

		ReportResultPage rp = pas.getReportPageWithTypedValues(reportData);
		PortalReport pReport = new PortalReport();
		List<PortalReportColumn> columnList = new ArrayList<>();
		List<PortalReportColumnData> dataList = new ArrayList<>();
		List<Long> identifierList = new ArrayList<>();

		// set portal report
		pReport.setName(processReport.getDisplay().getName());
		pReport.setDescription(processReport.getDisplay().getDescription());

		// set columns
		for (Column column : reportData.getColumns()) {
			if (column.getShow()) {
				PortalReportColumn pColumn = new PortalReportColumn();

				// unescape characters for the column label (e.g # & < >)
				pColumn.setLabel(StringEscapeUtils.unescapeHtml(column.getName()));
				pColumn.setField(column.getLocalId().toString());
				pColumn.setAlignment((column.getFormatToken().equals("number")) ? "RIGHT" : "LEFT");

				columnList.add(pColumn);
				dataList.add(new PortalReportColumnData());
			}
		}

		// set data
		for (HashMap<String, Object> map : (HashMap<String, Object>[]) rp.getResults()) {
			identifierList.add((Long) ((TypedValue) map.get("id")).getValue());

			for (int i=0; i<columnList.size(); i++) {
				PortalReportColumn pColumn = columnList.get(i);
				PortalReportColumnData pData = dataList.get(i);
				PortalReportCell cell = new PortalReportCell();

				cell.setValue((TypedValue) map.get("c" + pColumn.getField()));
				cell.setDrilldown((TypedValue) map.get("dp" + pColumn.getField()));

				pData.addCell(cell);
			}
		}

		return new PortalReportDataSubset(
			pagingInfo,
			(int) rp.getAvailableItems(),
			dataList,
			identifierList,
			pReport,
			columnList
		);
	}

	private ReportData getReportDataWithContext(ServiceContext sc, ProcessReport processReport, Long[] contextIds, PortalReportFilter[] filters) {
		ReportData reportData = processReport.getData();
		int contextType = reportData.getContextType();

		switch(contextType) {
			case ReportData.CONTEXT_TYPE_ATTRIBUTED_TO_USER:
			case ReportData.CONTEXT_TYPE_TASK_OWNER:
			case ReportData.CONTEXT_TYPE_INITIATED_BY:
				String username = sc.getIdentity().getIdentity();

				LocalObject userContext = new LocalObject();
				userContext.setType(ObjectTypeMapping.TYPE_USER);
				userContext.setStringId(username);

				reportData.setContext(new LocalObject[] { userContext });
			break;
			default:
				if (contextIds != null && contextIds.length > 0) {
					LocalObject[] contexts = new LocalObject[contextIds.length];

					for (int i=0; i<contextIds.length; i++) {
						contexts[i] = new LocalObject();

						switch(contextType) {
							case ReportData.CONTEXT_TYPE_PROCESS_MODEL:
								contexts[i].setType(ObjectTypeMapping.TYPE_BPM_PROCESS_MODEL);
							break;
							case ReportData.CONTEXT_TYPE_PROCESS:
								contexts[i].setType(ObjectTypeMapping.TYPE_BPM_PROCESS);
							break;
							default:
								LOG.warn("Report context type " + contextType + " not supported");
							break;
						}

						contexts[i].setId(contextIds[i]);
					}

					reportData.setContext(contexts);
				}
			break;
		}

		reportData.setBaseFilters(getReportFilter(reportData, filters));

		return reportData;
	}

	private Filter[] getReportFilter(ReportData reportData, PortalReportFilter[] filters) {
		Filter[] baseFilters = reportData.getBaseFilters();

		if (filters == null || filters.length == 0) {
			return baseFilters;
		}

		List<Filter> filterList = new ArrayList<>(Arrays.asList(baseFilters));

		for (PortalReportFilter filter : filters) {
			Integer columnLocalId = Integer.parseInt(filter.getField());

			NamedTypedValue ntv = new NamedTypedValue(filter.getValue());

			SimpleColumnFilter baseFilter = new SimpleColumnFilter(
				columnLocalId,
				SimpleColumnFilter.ComparisonType.valueOf(filter.getOperator()),
				ntv,
				true
			);

			filterList.add(baseFilter);
		}

		return filterList.toArray(new Filter[filterList.size()]);
	}
}
